import os

SERVICENOW_URL = os.environ["SERVICENOW_URL"]
SERVICENOW_API_TOKEN = os.environ.get("SERVICENOW_API_TOKEN")

SERVICENOW_API_USER = os.environ.get("SERVICENOW_API_USER")
SERVICENOW_API_PASSWORD = os.environ.get("SERVICENOW_API_PASSWORD")
